import ipaddress
import re

PROPERTIES_FLAG = ["is_protected", "updated_at"]
PROPERTIES_OBJECT = ["source", "owner"]
SAFE_VALUE = re.compile(r"(^[\. /:a-zA-Z0-9_-]+$)|(^$)")

IP_TYPES = ipaddress.IPv4Interface | ipaddress.IPv6Interface | ipaddress.IPv4Network | ipaddress.IPv6Network

ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling artifact_fetch is only supported for nodes that are Artifact Definition target"
)
ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling artifact_generate is only supported for nodes that are Artifact Definition targets"
)
ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling generate is only supported for CoreArtifactDefinition nodes"
)

HIERARCHY_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE = "Hierarchical fields are not supported for this node."

HFID_STR_SEPARATOR = "__"



import ujson
from prefect import Flow

from infrahub.log import get_logger
from infrahub.message_bus import RPCErrorResponse, messages
from infrahub.message_bus.operations import git, refresh, send
from infrahub.message_bus.types import MessageTTL
from infrahub.services.adapters.message_bus import InfrahubMessageBus
from infrahub.tasks.check import set_check_status

COMMAND_MAP = {
    "git.file.get": git.file.get,
    "git.repository.connectivity": git.repository.connectivity,
    "refresh.git.fetch": git.repository.fetch,
    "refresh.registry.branches": refresh.registry.branches,
    "refresh.registry.rebased_branch": refresh.registry.rebased_branch,
    "send.echo.request": send.echo.request,
}


async def execute_message(
    routing_key: str, message_body: bytes, message_bus: InfrahubMessageBus, skip_flow: bool = False
) -> MessageTTL | None:
    message_data = ujson.loads(message_body)
    message = messages.MESSAGE_MAP[routing_key](**message_data)
    message.set_log_data(routing_key=routing_key)
    try:
        func = COMMAND_MAP[routing_key]
        if skip_flow and isinstance(func, Flow):
            func = func.fn
        await func(message=message)
    except Exception as exc:
        if message.reply_requested:
            response = RPCErrorResponse(errors=[str(exc)], initial_message=message.model_dump())
            await message_bus.reply_if_initiator_meta(message=response, initiator=message)
            return None
        if message.reached_max_retries:
            get_logger().exception("Message failed after maximum number of retries", error=exc)
            await set_check_status(message, conclusion="failure")
            return None
        message.increase_retry_count()
        await message_bus.send(message, delay=MessageTTL.FIVE, is_retry=True)
        return MessageTTL.FIVE



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import ID, Field, Int, List, NonNull, String

from infrahub.core.registry import registry
from infrahub.exceptions import ValidationError
from infrahub.graphql.field_extractor import extract_graphql_fields
from infrahub.graphql.types import BranchType, InfrahubBranch, InfrahubBranchType

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo


async def branch_resolver(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    **kwargs: Any,
) -> list[dict[str, Any]]:
    fields = extract_graphql_fields(info)
    return await BranchType.get_list(graphql_context=info.context, fields=fields, exclude_global=True, **kwargs)


BranchQueryList = Field(
    List(of_type=NonNull(BranchType)),
    ids=List(ID),
    name=String(),
    description="Retrieve information about active branches.",
    resolver=branch_resolver,
    required=True,
)


async def infrahub_branch_resolver(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    limit: int | None = None,
    offset: int | None = None,
    name__value: str | None = None,
    ids: list[str] | None = None,
) -> dict[str, Any]:
    if isinstance(limit, int) and limit < 1:
        raise ValidationError("limit must be >= 1")
    if isinstance(offset, int) and offset < 0:
        raise ValidationError("offset must be >= 0")

    fields = extract_graphql_fields(info)
    result: dict[str, Any] = {}
    if "edges" in fields:
        branches = await InfrahubBranch.get_list(
            graphql_context=info.context,
            fields=fields.get("edges", {}).get("node", {}),
            limit=limit,
            offset=offset,
            name=name__value,
            ids=ids,
            exclude_global=True,
        )
        result["edges"] = [{"node": branch} for branch in branches]
    if "count" in fields:
        result["count"] = await InfrahubBranchType.get_list_count(
            graphql_context=info.context, name=name__value, ids=ids
        )

    if "default_branch" in fields:
        result["default_branch"] = await InfrahubBranch.get_by_name(
            graphql_context=info.context,
            fields=fields["default_branch"],
            name=registry.default_branch,
        )

    return result


InfrahubBranchQueryList = Field(
    InfrahubBranchType,
    offset=Int(),
    limit=Int(),
    name__value=String(),
    ids=List(ID),
    description="Retrieve paginated information about active branches.",
    resolver=infrahub_branch_resolver,
    required=True,
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, Field, Int, List, NonNull, String

from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME

from ...exceptions import BranchNotFoundError
from .enums import InfrahubBranchStatus
from .standard_node import InfrahubObjectType

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext


class BranchType(InfrahubObjectType):
    id = String(required=True)
    name = String(required=True)
    description = String(required=False)
    origin_branch = String(required=False)
    branched_from = String(required=False)
    status = InfrahubBranchStatus(required=True)
    graph_version = Int(required=False)
    created_at = String(required=False)
    sync_with_git = Boolean(required=False)
    is_default = Boolean(required=False)
    is_isolated = Field(Boolean(required=False), deprecation_reason="non isolated mode is not supported anymore")
    has_schema_changes = Boolean(required=False)

    class Meta:
        description = "Branch"
        name = "Branch"
        model = Branch

    @staticmethod
    async def _map_fields_to_graphql(objs: list[Branch], fields: dict) -> list[dict[str, Any]]:
        return [await obj.to_graphql(fields=fields) for obj in objs]

    @classmethod
    async def get_list(
        cls,
        fields: dict,
        graphql_context: GraphqlContext,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        async with graphql_context.db.start_session(read_only=True) as db:
            objs = await Branch.get_list(db=db, **kwargs)

            if not objs:
                return []

            return await cls._map_fields_to_graphql(objs=objs, fields=fields)

    @classmethod
    async def get_by_name(
        cls,
        fields: dict,
        graphql_context: GraphqlContext,
        name: str,
    ) -> dict[str, Any]:
        branch_responses = await cls.get_list(fields=fields, graphql_context=graphql_context, name=name)

        if branch_responses:
            return branch_responses[0]
        raise BranchNotFoundError(f"Branch with name '{name}' not found")


class RequiredStringValueField(InfrahubObjectType):
    value = String(required=True)


class NonRequiredStringValueField(InfrahubObjectType):
    value = String(required=False)


class NonRequiredIntValueField(InfrahubObjectType):
    value = Int(required=False)


class NonRequiredBooleanValueField(InfrahubObjectType):
    value = Boolean(required=False)


class StatusField(InfrahubObjectType):
    value = InfrahubBranchStatus(required=True)


class InfrahubBranch(BranchType):
    name = Field(RequiredStringValueField, required=True)
    description = Field(NonRequiredStringValueField, required=False)
    origin_branch = Field(NonRequiredStringValueField, required=False)
    branched_from = Field(NonRequiredStringValueField, required=False)
    graph_version = Field(NonRequiredIntValueField, required=False)
    status = Field(StatusField, required=True)
    sync_with_git = Field(NonRequiredBooleanValueField, required=False)
    is_default = Field(NonRequiredBooleanValueField, required=False)
    is_isolated = Field(
        NonRequiredBooleanValueField, required=False, deprecation_reason="non isolated mode is not supported anymore"
    )
    has_schema_changes = Field(NonRequiredBooleanValueField, required=False)

    class Meta:
        description = "InfrahubBranch"
        name = "InfrahubBranch"

    @staticmethod
    async def _map_fields_to_graphql(objs: list[Branch], fields: dict) -> list[dict[str, Any]]:
        field_keys = fields.keys()
        result: list[dict[str, Any]] = []
        for obj in objs:
            if obj.name == GLOBAL_BRANCH_NAME:
                continue
            data: dict[str, Any] = {}
            for field in field_keys:
                if field == "id":
                    data["id"] = obj.uuid
                    continue
                value = getattr(obj, field, None)
                if isinstance(fields.get(field), dict):
                    data[field] = {"value": value}
                else:
                    data[field] = value
            result.append(data)
        return result


class InfrahubBranchEdge(InfrahubObjectType):
    node = Field(InfrahubBranch, required=True)


class InfrahubBranchType(InfrahubObjectType):
    count = Field(Int, description="Total number of items")
    edges = Field(NonNull(List(of_type=NonNull(InfrahubBranchEdge))))
    default_branch = Field(
        InfrahubBranch,
        required=True,
        description="The default branch of the Infrahub instance, provides a direct way to access the default branch regardless of filters.",
    )

    @classmethod
    async def get_list_count(cls, graphql_context: GraphqlContext, **kwargs: Any) -> int:
        async with graphql_context.db.start_session(read_only=True) as db:
            return await Branch.get_list_count(db=db, **kwargs)



from typing import Any

from infrahub.core.attribute import BaseAttribute
from infrahub.core.branch import Branch
from infrahub.core.node import Node
from infrahub.database import InfrahubDatabase

from .queries.get_profile_data import GetProfileDataQuery, ProfileData


class NodeProfilesApplier:
    """Applies profile values to nodes and templates.

    Profile values take precedence over both default values and template-sourced values.
    When a template has profiles assigned:
    1. Profile values are applied to the template itself
    2. Nodes created from that template inherit the profile values (not the template's own values)
    3. Profile priority determines which profile wins when multiple profiles set the same attribute
    """

    def __init__(self, db: InfrahubDatabase, branch: Branch):
        self.db = db
        self.branch = branch

    async def _get_profile_ids(self, node: Node) -> list[str]:
        try:
            profiles_rel = node.get_relationship("profiles")
        except ValueError:
            return []
        profile_rels = await profiles_rel.get_relationships(db=self.db)
        return [pr.peer_id for pr in profile_rels if pr.peer_id]

    async def _get_attr_names_for_profiles(self, node: Node) -> list[str]:
        node_schema = node.get_schema()

        # get the names of attributes that could be affected by profile changes
        attr_names_for_profiles: list[str] = []
        for attr_schema in node_schema.attributes:
            attr_name = attr_schema.name
            node_attr: BaseAttribute = getattr(node, attr_name)
            if node_attr.is_from_profile or node_attr.is_default:
                attr_names_for_profiles.append(attr_name)
        return attr_names_for_profiles

    async def _get_sorted_profile_data(
        self, profile_ids: list[str], attr_names_for_profiles: list[str]
    ) -> list[ProfileData]:
        if not profile_ids:
            return []
        query = await GetProfileDataQuery.init(
            db=self.db, branch=self.branch, profile_ids=profile_ids, attr_names=attr_names_for_profiles
        )
        await query.execute(db=self.db)
        profile_data_list = query.get_profile_data()
        return sorted(profile_data_list, key=lambda x: (x.priority, x.uuid))

    def _apply_profile_to_attribute(self, node_attr: BaseAttribute, profile_value: Any, profile_id: str) -> bool:
        is_changed = False
        if node_attr.value != profile_value:
            node_attr.value = profile_value
            is_changed = True
        if node_attr.is_default is not False:
            node_attr.is_default = False
            is_changed = True
        if node_attr.is_from_profile is not True:
            node_attr.is_from_profile = True
            is_changed = True
        if node_attr.source_id != profile_id:  # type: ignore[attr-defined]
            node_attr.set_source(value=profile_id)
            is_changed = True
        return is_changed

    def _remove_profile_from_attribute(self, node_attr: BaseAttribute) -> None:
        node_attr.clear_source()
        node_attr.value = node_attr.schema.default_value
        node_attr.is_default = True
        node_attr.is_from_profile = False

    async def apply_profiles(self, node: Node) -> list[str]:
        profile_ids = await self._get_profile_ids(node=node)
        attr_names_for_profiles = await self._get_attr_names_for_profiles(node=node)

        if not attr_names_for_profiles:
            return []

        # get profiles priorities and attribute values on branch
        sorted_profile_data = await self._get_sorted_profile_data(
            profile_ids=profile_ids, attr_names_for_profiles=attr_names_for_profiles
        )

        updated_field_names = []
        # set attribute values/is_default/is_from_profile on nodes
        for attr_name in attr_names_for_profiles:
            has_profile_data = False
            node_attr = node.get_attribute(attr_name)
            for profile_data in sorted_profile_data:
                profile_value = profile_data.attribute_values.get(attr_name)
                if profile_value is not None:
                    has_profile_data = True
                    is_changed = False
                    is_changed = self._apply_profile_to_attribute(
                        node_attr=node_attr, profile_value=profile_value, profile_id=profile_data.uuid
                    )
                    if is_changed:
                        updated_field_names.append(attr_name)
                    break
            if not has_profile_data and node_attr.is_from_profile:
                self._remove_profile_from_attribute(node_attr=node_attr)
                updated_field_names.append(attr_name)
        return updated_field_names



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Self, cast

import httpx
from graphene import Boolean, Field, InputObjectType, Mutation, String

from infrahub import config
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.schema import NodeSchema
from infrahub.git.models import (
    GitRepositoryImportObjects,
    GitRepositoryPullReadOnly,
)
from infrahub.graphql.types.common import IdentifierInput
from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.message_bus.messages.git_repository_connectivity import GitRepositoryConnectivityResponse
from infrahub.repositories.create_repository import RepositoryFinalizer
from infrahub.workflows.catalogue import (
    GIT_REPOSITORIES_IMPORT_OBJECTS,
    GIT_REPOSITORIES_PULL_READ_ONLY,
)

from ...core.node.create import create_node
from ..types.task import TaskInfo
from .main import InfrahubMutationMixin, InfrahubMutationOptions, build_graphql_response

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.protocols import CoreReadOnlyRepository, CoreRepository
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext

log = get_logger()


class InfrahubRepositoryMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(cls, schema: NodeSchema | None = None, _meta=None, **options) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        override_data: dict[str, Any] | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        cleanup_payload(data)
        db = database or graphql_context.db
        create_data = dict(data)
        create_data.update(override_data or {})
        obj = await create_node(data=create_data, db=db, branch=branch, schema=cls._meta.active_schema)

        await RepositoryFinalizer(
            account_session=graphql_context.active_account_session,
            services=graphql_context.active_service,
            context=graphql_context.get_context(),
        ).post_create(
            obj=obj,  # type: ignore
            branch=branch,
            db=db,
        )

        graphql_response = await build_graphql_response(info=info, db=db, obj=obj)
        return obj, cls(**graphql_response)

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        cleanup_payload(data)
        if not node:
            node: CoreReadOnlyRepository | CoreRepository = await NodeManager.get_one_by_id_or_default_filter(
                db=graphql_context.db,
                kind=cls._meta.schema.kind,
                id=data.get("id"),
                branch=branch,
                include_owner=True,
                include_source=True,
            )
        if node.get_kind() != InfrahubKind.READONLYREPOSITORY:
            return await super().mutate_update(info, data, branch, database=graphql_context.db, node=node)

        node = cast("CoreReadOnlyRepository", node)
        current_commit = node.commit.value
        current_ref = node.ref.value
        new_commit = None
        if data.commit and data.commit.value:
            new_commit = data.commit.value
        new_ref = None
        if data.ref and data.ref.value:
            new_ref = data.ref.value

        obj, result = await super().mutate_update(info, data, branch, database=graphql_context.db, node=node)
        obj = cast("CoreReadOnlyRepository", obj)

        send_update_message = (new_commit and new_commit != current_commit) or (new_ref and new_ref != current_ref)
        if not send_update_message:
            return obj, result

        log.info(
            "update read-only repository commit",
            name=obj.name.value,
            commit=data.commit.value if data.commit else None,
            ref=data.ref.value if data.ref else None,
        )

        model = GitRepositoryPullReadOnly(
            repository_id=obj.id,
            repository_name=obj.name.value,
            location=obj.location.value,
            ref=obj.ref.value,
            commit=new_commit,
            infrahub_branch_name=branch.name,
            infrahub_branch_id=str(branch.get_uuid()),
        )
        if graphql_context.service:
            await graphql_context.service.workflow.submit_workflow(
                workflow=GIT_REPOSITORIES_PULL_READ_ONLY,
                context=graphql_context.get_context(),
                parameters={"model": model},
            )
        return obj, result


def cleanup_payload(data: InputObjectType | dict[str, Any]) -> None:
    """If the input payload contains an http URL that doesn't end in .git it will be added to the payload"""
    http_without_dotgit = r"^(https?://)(?!.*\.git$).*"
    if (
        data.get("location")
        and data["location"].get("value")
        and re.match(http_without_dotgit, data["location"]["value"])
    ):
        url = httpx.URL(data["location"]["value"])
        if url.host in config.SETTINGS.git.append_git_suffix:
            data["location"]["value"] = f"{data['location']['value']}.git"


class ProcessRepository(Mutation):
    class Arguments:
        data = IdentifierInput(required=True)

    ok = Boolean()
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: IdentifierInput,
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context
        branch = graphql_context.branch
        repository_id = str(data.id)
        repo: CoreReadOnlyRepository | CoreRepository = await NodeManager.get_one_by_id_or_default_filter(
            db=graphql_context.db,
            kind=InfrahubKind.GENERICREPOSITORY,
            id=str(data.id),
            branch=branch,
        )

        model = GitRepositoryImportObjects(
            repository_id=repository_id,
            repository_name=str(repo.name.value),
            repository_kind=repo.get_kind(),
            commit=str(repo.commit.value),
            infrahub_branch_name=branch.name,
        )
        workflow = await graphql_context.active_service.workflow.submit_workflow(
            workflow=GIT_REPOSITORIES_IMPORT_OBJECTS, context=graphql_context.get_context(), parameters={"model": model}
        )
        task = {"id": workflow.id}
        return cls(ok=True, task=task)


class ValidateRepositoryConnectivity(Mutation):
    class Arguments:
        data = IdentifierInput(required=True)

    ok = Boolean(required=True)
    message = String(required=True)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: IdentifierInput,
    ) -> dict[str, Any]:
        graphql_context: GraphqlContext = info.context
        branch = graphql_context.branch
        repository_id = str(data.id)
        repo: CoreReadOnlyRepository | CoreRepository = await NodeManager.get_one_by_id_or_default_filter(
            db=graphql_context.db,
            kind=InfrahubKind.GENERICREPOSITORY,
            id=repository_id,
            branch=branch,
        )

        message = messages.GitRepositoryConnectivity(
            repository_name=str(repo.name.value),
            repository_location=str(repo.location.value),
        )
        if graphql_context.service:
            response = await graphql_context.service.message_bus.rpc(
                message=message, response_class=GitRepositoryConnectivityResponse
            )

        return {"ok": response.data.success, "message": response.data.message}



from __future__ import annotations

from datetime import timedelta
from enum import StrEnum
from typing import TYPE_CHECKING, Any, TypeVar

from prefect.events.actions import RunDeployment
from prefect.events.schemas.automations import Automation, Posture
from prefect.events.schemas.automations import EventTrigger as PrefectEventTrigger
from prefect.events.schemas.events import ResourceSpecification
from pydantic import BaseModel, Field

from infrahub import __version__
from infrahub.workflows.models import WorkflowDefinition  # noqa: TC001

from .constants import NAME_SEPARATOR

if TYPE_CHECKING:
    from uuid import UUID

T = TypeVar("T", bound="TriggerDefinition")


class TriggerComparison(StrEnum):
    MATCH = "match"  # Expected trigger and actual trigger is identical
    REFRESH = "refresh"  # The branch parameters doesn't match, the hash does, refresh in Prefect but don't run triggers
    UPDATE = "update"  # Neither branch or other data points match, update in Prefect and run triggers

    @property
    def update_prefect(self) -> bool:
        return self in {TriggerComparison.REFRESH, TriggerComparison.UPDATE}


class TriggerSetupReport(BaseModel):
    created: list[TriggerDefinition] = Field(default_factory=list)
    refreshed: list[TriggerDefinition] = Field(default_factory=list)
    updated: list[TriggerDefinition] = Field(default_factory=list)
    deleted: list[Automation] = Field(default_factory=list)
    unchanged: list[TriggerDefinition] = Field(default_factory=list)

    @property
    def in_use_count(self) -> int:
        return len(self.created + self.updated + self.unchanged + self.refreshed)

    def add_with_comparison(self, trigger: TriggerDefinition, comparison: TriggerComparison) -> None:
        match comparison:
            case TriggerComparison.UPDATE:
                self.updated.append(trigger)
            case TriggerComparison.REFRESH:
                self.refreshed.append(trigger)
            case TriggerComparison.MATCH:
                self.unchanged.append(trigger)

    def _created_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        return [trigger for trigger in self.created if isinstance(trigger, trigger_type)]

    def _refreshed_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        return [trigger for trigger in self.refreshed if isinstance(trigger, trigger_type)]

    def _unchanged_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        return [trigger for trigger in self.unchanged if isinstance(trigger, trigger_type)]

    def _updated_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        return [trigger for trigger in self.updated if isinstance(trigger, trigger_type)]

    def triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        """Return all triggers that match the specified type.

        Args:
            trigger_type: A TriggerDefinition class or subclass to filter by

        Returns:
            List of triggers of the specified type from all categories
        """
        created = self._created_triggers_with_type(trigger_type=trigger_type)
        updated = self._updated_triggers_with_type(trigger_type=trigger_type)
        refreshed = self._refreshed_triggers_with_type(trigger_type=trigger_type)
        unchanged = self._unchanged_triggers_with_type(trigger_type=trigger_type)
        return created + updated + refreshed + unchanged

    def modified_triggers_with_type(self, trigger_type: type[T]) -> list[T]:
        """Return all created and updated triggers that match the specified type.

        Args:
            trigger_type: A TriggerDefinition class or subclass to filter by

        Returns:
            List of triggers of the specified type from both created and updated lists
        """
        created = self._created_triggers_with_type(trigger_type=trigger_type)
        updated = self._updated_triggers_with_type(trigger_type=trigger_type)
        return created + updated


class TriggerType(StrEnum):
    ACTION_TRIGGER_RULE = "action_trigger_rule"
    BUILTIN = "builtin"
    WEBHOOK = "webhook"
    COMPUTED_ATTR_JINJA2 = "computed_attr_jinja2"
    COMPUTED_ATTR_PYTHON = "computed_attr_python"
    COMPUTED_ATTR_PYTHON_QUERY = "computed_attr_python_query"
    DISPLAY_LABEL_JINJA2 = "display_label_jinja2"
    HUMAN_FRIENDLY_ID = "human_friendly_id"
    # OBJECT = "object"

    @property
    def is_branch_specific(self) -> bool:
        return self in {
            TriggerType.COMPUTED_ATTR_JINJA2,
            TriggerType.COMPUTED_ATTR_PYTHON,
            TriggerType.COMPUTED_ATTR_PYTHON_QUERY,
            TriggerType.DISPLAY_LABEL_JINJA2,
            TriggerType.HUMAN_FRIENDLY_ID,
        }


def _match_related_dict() -> dict:
    # Make Mypy happy as match related is a dict[str, Any] | list[dict[str, Any]]
    return {}


class EventTrigger(BaseModel):
    events: set = Field(default_factory=set)
    match: dict[str, Any] = Field(default_factory=dict)
    match_related: dict[str, Any] | list[dict[str, Any]] = Field(default_factory=_match_related_dict)

    def get_prefect(self) -> PrefectEventTrigger:
        return PrefectEventTrigger(
            posture=Posture.Reactive,
            expect=self.events,
            within=timedelta(0),
            match=ResourceSpecification(self.match),
            match_related=self.related_resource_specification,
            threshold=1,
        )

    @property
    def related_resource_specification(self) -> ResourceSpecification | list[ResourceSpecification]:
        if isinstance(self.match_related, dict):
            return ResourceSpecification(self.match_related)

        if len(self.match_related) == 1:
            return ResourceSpecification(self.match_related[0])

        return [ResourceSpecification(related_match) for related_match in self.match_related]


class ExecuteWorkflow(BaseModel):
    workflow: WorkflowDefinition
    parameters: dict[str, Any] = Field(default_factory=dict)

    @property
    def name(self) -> str:
        return self.workflow.name

    def get_prefect(self, mapping: dict[str, UUID]) -> RunDeployment:
        deployment_id = mapping[self.name]
        return self.get(deployment_id)

    def get(self, id: UUID) -> RunDeployment:
        return RunDeployment(
            source="selected",
            deployment_id=id,
            parameters=self.parameters,
            job_variables={},
        )

    def validate_parameters(self) -> None:
        if not self.parameters:
            return

        workflow_params = self.workflow.get_parameters()
        workflow_required_params = [p.name for p in workflow_params.values() if p.required]
        trigger_params = list(self.parameters.keys())

        missing_required_params = set(workflow_required_params) - set(trigger_params)
        wrong_params = set(trigger_params) - set(workflow_params)

        if missing_required_params:
            raise ValueError(
                f"Missing required parameters: {missing_required_params} for workflow {self.workflow.name}"
            )

        if wrong_params:
            raise ValueError(f"Workflow {self.workflow.name} doesn't support parameters: {wrong_params}")


class TriggerDefinition(BaseModel):
    name: str
    type: TriggerType
    previous_names: set = Field(default_factory=set)
    description: str = ""
    trigger: EventTrigger
    actions: list[ExecuteWorkflow]

    def get_deployment_names(self) -> list[str]:
        """Return the name of all deployments used by this trigger"""
        return [action.name for action in self.actions]

    def get_description(self) -> str:
        return f"Automation for Trigger {self.name} of type {self.type.value} (v{__version__})"

    def generate_name(self) -> str:
        return f"{self.type.value}{NAME_SEPARATOR}{self.name}"

    def validate_actions(self) -> None:
        for action in self.actions:
            action.validate_parameters()


class TriggerBranchDefinition(TriggerDefinition):
    branch: str

    def generate_name(self) -> str:
        return f"{self.type.value}{NAME_SEPARATOR}{self.branch}{NAME_SEPARATOR}{self.name}"


class BuiltinTriggerDefinition(TriggerDefinition):
    type: TriggerType = TriggerType.BUILTIN
